#include "scnhacking.h"

entity_t **alloc_entities(unsigned int n) {
  return (entity_t **) alloc_lots(n, "entity", sizeof(entity_t *), (void *(*)()) alloc_entity);
}

entity_t *alloc_entity() {
  return (entity_t *) alloc_one("entity", sizeof(entity_t));
}

void cleanup_entities(thing_t *entities) {
  int i;
  entity_t *entity;

  for (i = 0; i < entities->num; i++) {
    entity = (entity_t *) entities->things[i];
    TRACE2("freeing entity at %p\n", entity);
    cleanup_entity(entity);
  }

  free(entities->things);
}

void cleanup_entity(entity_t *entity) {
  if (! entity) {
    ERROR("no entity!\n");
    return;
  }

  cleanup_pairs(entity->num_fields, entity->fields);
  TRACE2("freeing entity at %p\n", entity);
  free(entity);
}

int find_num_entities(scn_t *scn) {
  struct thing *entities;
  entity_t **new_entities;

  if (! scn || ! scn->fd) {
    ERROR("no SCN!\n");
    return 1;
  }

  entities = &scn->entities;

  /* skip to magic offset */
  if (fseek(scn->fd, ENTITIES_NUM_OFFSET, SEEK_SET)) {
    ERROR("can't seek to location of entities offset in %s\n", scn->filename);
    return 2;
  }

  /* read number of entities */
  if (read_uint32(scn->fd, (unsigned int *) &entities->num)) {
    ERROR("can't read number of entities in %s\n", scn->filename);
    return 3;
  }
  TRACE("%s: entities: %u\n", scn->filename, entities->num);

  /* skip to magic offset */
  if (fseek(scn->fd, ENTITIES_START_OFFSET, SEEK_SET)) {
    ERROR("can't seek to location of entities offset in %s\n", scn->filename);
    return 2;
  }

  /* read start of entities */
  if (read_uint32(scn->fd, (unsigned int *) &entities->offset)) {
    ERROR("can't read entities offset in %s\n", scn->filename);
    return 3;
  }
  TRACE("%s: entities offset: 0x%08x\n", scn->filename, entities->offset);
  /* read data stream length */
  if (read_uint32(scn->fd, (unsigned int *) &entities->length)) {
    ERROR("can't read entities data length in %s\n", scn->filename);
    return 4;
  }
  TRACE("%s: entities length: 0x%08x\n", scn->filename, entities->length);
  /* remember initial length because it might change if we edit stuff */
  entities->orig_length = entities->length;
  DEBUG("%s: entities: offset: 0x%08x, number: %u, length: 0x%04x\n", scn->filename, entities->offset, entities->num, entities->length);

  /* allocate them */
  new_entities = alloc_entities(entities->num);
  if (! new_entities) {
    ERROR("out of memory for entities");
    entities->num = 0;
    return 5;
  }
  entities->things = (void **) new_entities;

  return 0;
}

int find_entities(scn_t *scn) {
  thing_t *entities;
  entity_t *entity;
  off_t pos;
  int i, n;

  if (! scn || ! scn->fd) {
    ERROR("no SCN!\n");
    return 1;
  }

  entities = &scn->entities;

  /* skip to offset */
  if (fseek(scn->fd, entities->offset, SEEK_SET)) {
    ERROR("can't seek to start of entity in %s\n", scn->filename);
    return 2;
  }

  /* n = light index */
  n = 0;
  for (i = 0; i < entities->num; i++) {
    pos = get_file_position(scn->fd);
    TRACE("file position for entity %u: 0x%08x\n", i, pos);
    entity = (entity_t *) entities->things[i];
    entity->offset = pos;
    entity->id = i;
    if (parse_entity(scn, entity, i, &n)) {
      ERROR("can't parse entity %u\n", i);
      return 3;
    }
    if (classify_entity(scn, entity)) {
      ERROR("can't classify entity %u\n", i);
      return 3;
    }
  }

  /* sanity */
  pos = get_file_position(scn->fd);
  if (pos != scn->lightmaps.offset) {
    WARN("entity data stream ends at offset 0x%08x (expected 0x%08x)\n", pos, entities->offset);
  }
  pos -= entities->offset;
  if (pos != entities->length) {
    WARN("entity data stream read %u doesn't match header %u\n", pos, entities->length);
  }

  /* sort entclasses */
  if (sort_entclasses(scn->num_entclasses, scn->entclasses)) {
    WARN("can't sort entclasses\n");
  }

  return 0;
}

int parse_entity(scn_t *scn, entity_t *entity, unsigned int num, unsigned int *n) {
  unsigned int i;
  pair_t *pair;

  if (! scn || ! scn->fd) {
    ERROR("no SCN!\n");
    return 1;
  }

  /* read number of information entities */
  if (read_uint32(scn->fd, &entity->num_fields)) {
    ERROR("can't read number of information fields\n");
    return 3;
  }
  TRACE("entity %u num fields: %u\n", num, entity->num_fields);
  /* read unknown */
  if (read_uint32(scn->fd, &entity->unknown)) {
    ERROR("can't read unknown thing\n");
    return 4;
  }
  TRACE("entity %u unknown: %u\n", num, entity->unknown);

  /* allocate space for them */
  entity->fields = alloc_pairs(entity->num_fields);
  if (! entity->fields) {
    ERROR("out of memory for information fields");
    entity->num_fields = 0;
    return 5;
  }

  /* read 'em in */
  for (i = 0; i < entity->num_fields; i++) {
    pair = entity->fields[i];
    /* offset */
    pair->offset = get_file_position(scn->fd);
    /* key length */
    if (read_uint16(scn->fd, &pair->key_length)) {
      ERROR("can't read entity %u key %u length\n", entity->id, i);
      return 6;
    }
    pair->key = (char *) malloc(pair->key_length);
    if (! pair->key) {
      ERROR("out of memory for entity %u key %u\n", entity->id, i);
      return 7;
    }
    /* value length */
    if (read_uint16(scn->fd, &pair->value_length)) {
      ERROR("can't read entity %u value %u length\n", entity->id, i);
      return 8;
    }
    pair->value = (char *) malloc(pair->value_length);
    if (! pair->value) {
      ERROR("out of memory for entity %u value %u\n", entity->id, i);
      return 9;
    }
    /* key */
    if (read_arbitrary(scn->fd, pair->key, pair->key_length)) {
      ERROR("can't read entity %u key %u\n", entity->id, i);
      return 10;
    }
    /* value */
    if (read_arbitrary(scn->fd, pair->value, pair->value_length)) {
      ERROR("can't read entity %u value %u\n", entity->id, i);
      return 11;
    }
    /* sanity */
    pair->key[pair->key_length - 1] = '\0';
    pair->value[pair->value_length - 1] = '\0';
    DEBUG("entity %u: %s -> \"%s\"\n", num, pair->key, pair->value);
  }

  /* mark it if it's a light */
  if (is_light(entity)) {
    if (*n >= scn->lightmaps.num_lights) {
      ERROR("light index %u out of range\n", *n);
      return 12;
    }
    TRACE("entity %u is light %u\n", i, *n);
    scn->lightmaps.addresses[*n] = (void *) entity;
    ++*n;
  }

  return 0;
}

/* maintain a count of how many different types of entities there are */
int classify_entity(scn_t *scn, entity_t *entity) {
  unsigned int num_entclasses;
  unsigned short len;
  char *classname;
  entclass_t *entclass, **entclasses, **new_entclasses;
  entity_t**new_addresses;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }
  if (! entity) {
    ERROR("no entity!\n");
    return 2;
  }

  classname = get_entity_value(entity, CLASSNAME);
  if (! classname) {
    WARN("can't classify entity - file %s is broken\n", scn->basename);
    return 0;
  }
  DEBUG("entity %u is of type \"%s\"\n", entity->id, classname);

  entclasses = scn->entclasses;
  num_entclasses = scn->num_entclasses;

  if (! *classname) {
    WARN("entity %u has null %s attribute\n", entity->id, CLASSNAME);
    classname = "<no classname>";
  }

  /* get existing record */
  entclass = get_entclass(num_entclasses, entclasses, classname);
  if (! entclass) {
    /* we must create one */
    DEBUG("creating new entclasses\n");
    new_entclasses = alloc_more_entclasses(1, num_entclasses, entclasses);
    if (! new_entclasses) {
      ERROR("out of memory for new entclasses");
      return 4;
    }
    scn->entclasses = entclasses = new_entclasses;
    scn->num_entclasses++;

    entclass = entclasses[num_entclasses++];

    TRACE("adding classname \"%s\"\n", classname);
    entclass->classname = alloc_name(classname, &len);
    if (! len) {
      ERROR("out of memory for new entclass classname \"%s\"", classname);
      free(entclass);
      return 5;
    }
  }
  
  /* increase number of addresses (alloc_more() is safe if there were none) */
  DEBUG("adding address %p\n", entity);
  new_addresses = alloc_more_addresses(1, entclass->num, entclass->addresses);
  if (! new_addresses) {
    ERROR("out of memory for addresses");
    free(entclass->classname);
    free(entclass);
    return 6;
  }

  entclass->addresses = new_addresses;
  /* chalk up another one */
  entclass->addresses[entclass->num++] = entity;

  return 0;
}

int write_entities(scn_t *scn) {
  int i;
  unsigned int j;
  thing_t *entities;
  entity_t *entity;
  pair_t *pair;
  unsigned int count;

  if (! scn || ! scn->out_fd) {
    ERROR("no SCN!\n");
    return 1;
  }

  entities = &scn->entities;

  DEBUG("writing %u entit%s\n", entities->num, (entities->num == 1) ? "y" : "ies");
  count = 0;
  for (i = 0; i < entities->num; i++) {
    entity = (entity_t *) entities->things[i];
    TRACE("writing entity %u\n", i);
    /* number of fields */
    if (write_uint32(scn->out_fd, &entity->num_fields)) {
      ERROR("can't write entity %u number of fields\n", i);
      return 2;
    }
    count += sizeof(entity->num_fields);
    /* unknown */
    if (write_uint32(scn->out_fd, &entity->unknown)) {
      ERROR("can't write entity %u unknown\n", i);
      return 3;
    }
    count += sizeof(entity->unknown);
    for (j = 0; j < entity->num_fields; j++) {
      pair = entity->fields[j];
      /* key length */
      if (write_uint16(scn->out_fd, &pair->key_length)) {
        ERROR("can't write entity %u key length %u\n", i, j);
        return 4;
      }
      count += sizeof(pair->key_length);
      /* value length */
      if (write_uint16(scn->out_fd, &pair->value_length)) {
        ERROR("can't write entity %u value length %u\n", i, j);
        return 5;
      }
      count += sizeof(pair->value_length);
      /* key */
      if (write_arbitrary(scn->out_fd, pair->key, pair->key_length)) {
        ERROR("can't write entity %u key %u\n", i, j);
        return 6;
      }
      count += pair->key_length;
      /* value */
      if (write_arbitrary(scn->out_fd, pair->value, pair->value_length)) {
        ERROR("can't write entity %u value %u\n", i, j);
        return 7;
      }
      count += pair->value_length;
      DEBUG("wrote entity %u: %s -> \"%s\"\n", i, pair->key, pair->value);
    }
    DEBUG("wrote entity %u - count is now %u\n", i, count);
  }

  if (count != entities->length) {
    WARN("entity data stream written %u doesn't match header %u\n", count, entities->length);
  }

  return 0;
}

char *get_entity_value(entity_t *entity, char *key) {
  unsigned int i;
  struct pair *pair;

  if (! entity) {
    ERROR("no entity!\n");
    return 0;
  }
  if (! key) {
    ERROR("no key!\n");
    return 0;
  }

  for (i = 0; i < entity->num_fields; i++) {
    pair = entity->fields[i];
    if (str_equiv(key, pair->key)) return pair->value;
  }

  ERROR("key \"%s\" not found in entity %u attribute set\n", key, entity->id);
  return 0;
}

entity_search_set_t *alloc_entity_search_set() {
  return (entity_search_set_t *) alloc_one("entity_search_set", sizeof(entity_search_set_t));
}

void cleanup_entity_search_sets(entity_search_set_t *sets) {
  entity_search_set_t *tmp;
  GLL_LOOP(tmp, sets) {
    if (tmp->key) free(tmp->key);
    if (tmp->value) free(tmp->value);
  }
  TRACE2("freeing entity_edit_sets at %p\n", sets);
  gll_free((gll_t *) sets);
}

int add_to_entity_search_set(entity_search_set_t **sets, char *entity_arg, char *key, char *value) {
  entity_search_set_t *last, *tmp;
  char *bogus;
  unsigned short len;

  tmp = alloc_entity_search_set();
  if (! tmp) {
    ERROR("out of memory for entity_search set");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);

  /* search by entity ID */
  if (entity_arg) {
    if (str_equiv(entity_arg, "any")) goto any_entity;
    tmp->entity_id = (unsigned int) strtol(entity_arg, &bogus, 0);
    if (*bogus) {
      ERROR("bogus entity ID \"%s\"- use a number\n", entity_arg);
      cleanup_entity_search_sets(tmp);
      return 2;
    }
    TRACE("tmp->entity_id: %u\n", tmp->entity_id);
  }
  else {
any_entity:
    tmp->entity_id = 0xffffffff;
    TRACE("tmp->entity_id: %u - any entity\n", tmp->entity_id);
  }

  /* search by key */
  if (key) {
    if (! *key) {
      ERROR("blank key name\n");
      cleanup_entity_search_sets(tmp);
      return 3;
    }
    tmp->key = alloc_name(key, &len);
    if (! len) {
      ERROR("out of memory for key name");
      return 4;
    }
  }
  else {
    tmp->key = 0;
    TRACE("tmp->key: %u - any key\n", tmp->key);
  }

  /* search by value */
  if (value) {
    if (! *value) goto blank_value;
    tmp->value = alloc_name(value, &len);
    if (! len) {
      ERROR("out of memory for value name");
      return 6;
    }
  }
  else {
blank_value:
    tmp->value = 0;
    TRACE("tmp->value: %u - any value\n", tmp->value);
  }

  last = (entity_search_set_t *) gll_last((gll_t *) *sets);
  TRACE("last: %p\n", last);
  /* this is the first one */
  if (last) last->next = tmp;
  else *sets = tmp;
  gll_debug((gll_t *) *sets);

  return 0;
}

int search_entities(scn_t *scn) {
  int i, count;
  entity_search_set_t *set;
  thing_t *entities;
  entity_t *entity;
  int notme;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  entities = &scn->entities;

  count = 0;
  for (i = 0; i < entities->num; i++) {
    entity = (entity_t *) entities->things[i];
    GLL_LOOP(set, scn->option->entity_searches) {
      notme = check_entity(entity, set);
      if (notme) break;
    }
    if (! notme) {
      count++;
      describe_entity(entity, 0);
    }
  }

  if (! count) {
    INFO("no matches\n");
    return 2;
  }

  SAY("found %u matching entit%s in %s\n", count, (count == 1) ? "y" : "ies", scn->filename);

  return 0;
}

/* check each of the conditions and return if they are not met */
int check_entity(entity_t *entity, entity_search_set_t *set) {
  unsigned int i;
  pair_t *pair;

  if (! entity) {
    ERROR("no entity!\n");
    return 1;
  }
  if (! set) {
    ERROR("no set!\n");
    return 2;
  }

  DEBUG("checking entity at %p for ID %u, key \"%s\" and value \"%s\" from set at %p\n", entity, set->entity_id, set->key, set->value, set);

  /* ID */
  if (entity->id != set->entity_id && set->entity_id != 0xffffffff) {
    TRACE("explicit ID %u doesn't match\n", set->entity_id);
    return -1;
  }

  /* key */
  if (set->key) {
    for (i = 0; i < entity->num_fields; i++) {
      pair = entity->fields[i];
      if (str_equiv(pair->key, set->key)) break;
      TRACE("entity %u attribute %u: key \"%s\" doesn't match requested key \"%s\"\n", entity->id, i, pair->key, set->key);
      pair = 0;
    }
    if (! pair) {
      TRACE("key \"%s\" not found in entity %u\n", set->key, entity->id);
      return -2;
    }
    
    /* value check on this key */
    if (set->value) {
      if (str_equiv(pair->value, set->value)) {
        DEBUG("bingo! %s -> \"%s\" for entity %u\n", set->key, set->value, entity->id);
        return 0;
      }
      else {
        TRACE("value \"%s\" not found in entity %u matching key \"%s\"\n", set->value, entity->id, set->key);
        return -3;
      }
    }
  }
  else {
    DEBUG("no key check\n");
  }
  
  /* value check on all keys */
  if (set->value) {
    for (i = 0; i < entity->num_fields; i++) {
      pair = entity->fields[i];
      if (str_equiv(pair->value, set->value)) break;
      TRACE("entity %u attribute %u: key \"%s\" doesn't match requested key \"%s\"\n", entity->id, i, pair->key, set->key);
      pair = 0;
    }
    if (! pair) {
      DEBUG("value \"%s\" not found for any key in entity %u\n", set->key, entity->id);
      return -4;
    }
  }
  else {
    DEBUG("no value check\n");
  }
 
  return 0;
}

entity_add_set_t *alloc_entity_add_set() {
  return (entity_add_set_t *) alloc_one("entity_add_set", sizeof(entity_add_set_t));
}

void cleanup_entity_add_sets(entity_add_set_t *sets) {
  entity_add_set_t *tmp;
  GLL_LOOP(tmp, sets) {
    if (tmp->name) free(tmp->name);
    if (tmp->classname) free(tmp->classname);
  }
  TRACE2("freeing entity_add_sets at %p\n", sets);
  gll_free((gll_t *) sets);
}

int add_to_entity_add_set(entity_add_set_t **sets, char *classname, char *name) {
  entity_add_set_t *last, *tmp;
  unsigned short len;

  tmp = alloc_entity_add_set();
  if (! tmp) {
    ERROR("out of memory for entity_add_set");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);

  /* name */
  tmp->name = alloc_name(name, &len);
  if (! len) {
    ERROR("out of memory for name");
    cleanup_entity_add_sets(tmp);
    return 2;
  }
  TRACE("tmp->name: %s\n", tmp->name);

  /* classname */
  tmp->classname = alloc_name(classname, &len);
  if (! len) {
    ERROR("out of memory for classname");
    cleanup_entity_add_sets(tmp);
    return 3;
  }
  TRACE("tmp->classname: %s\n", tmp->classname);

  last = (entity_add_set_t *) gll_last((gll_t *) *sets);
  TRACE("last: %p\n", last);
  /* this is the first one */
  if (last) last->next = tmp;
  else *sets = tmp;
  /*gll_debug((gll_t *) *sets);*/

  return 0;
}

entity_delete_set_t *alloc_entity_delete_set() {
  return (entity_delete_set_t *) alloc_one("entity_delete_set", sizeof(entity_delete_set_t));
}

void cleanup_entity_delete_sets(entity_delete_set_t *sets) {
  entity_delete_set_t *tmp;
  GLL_LOOP(tmp, sets) {
    if (tmp->entity_arg) free(tmp->entity_arg);
  }
  TRACE2("freeing entity_delete_sets at %p\n", sets);
  gll_free((gll_t *) sets);
}

int add_to_entity_delete_set(entity_delete_set_t **sets, char *entity_arg) {
  entity_delete_set_t *last, *tmp;
  char *bogus;

  tmp = alloc_entity_delete_set();
  if (! tmp) {
    ERROR("out of memory for entity_delete_set");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);

  /* entity */
  tmp->entity_id = (unsigned int) strtol(entity_arg, &bogus, 0);
  if (*bogus) {
    ERROR("bogus entity ID \"%s\" - use a number\n", entity_arg);
    cleanup_entity_delete_sets(tmp);
    return 2;
  }
  TRACE("tmp->entity_id: %u\n", tmp->entity_id);

  last = (entity_delete_set_t *) gll_last((gll_t *) *sets);
  TRACE("last: %p\n", last);
  /* this is the first one */
  if (last) last->next = tmp;
  else *sets = tmp;
  /*gll_debug((gll_t *) *sets);*/

  return 0;
}

entity_edit_set_t *alloc_entity_edit_set() {
  return (entity_edit_set_t *) alloc_one("entity_edit_set", sizeof(entity_edit_set_t));
}

void cleanup_entity_edit_sets(entity_edit_set_t *sets) {
  entity_edit_set_t *tmp;
  GLL_LOOP(tmp, sets) {
    if (tmp->entity_arg) free(tmp->entity_arg);
    if (tmp->key) free(tmp->key);
    /*if (tmp->value) free(tmp->value);*/ /* XXX: already freed! */
  }
  TRACE2("freeing entity_edit_sets at %p\n", sets);
  gll_free((gll_t *) sets);
}

int add_to_entity_edit_set(entity_edit_set_t **sets, char *entity_arg, char *key, char *value, int options) {
  entity_edit_set_t *last, *tmp;
  char *bogus;
  unsigned short len;

  tmp = alloc_entity_edit_set();
  if (! tmp) {
    ERROR("out of memory for entity_edit_set");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);

  /* entity */
  tmp->entity_id = (unsigned int) strtol(entity_arg, &bogus, 0);
  if (*bogus) {
    ERROR("bogus entity ID \"%s\" - use a number\n", entity_arg);
    cleanup_entity_edit_sets(tmp);
    return 2;
  }
  TRACE("tmp->entity_id: %u\n", tmp->entity_id);

  /* key */
  if (! *key) {
    ERROR("blank key name\n");
    cleanup_entity_edit_sets(tmp);
    return 3;
  }
  tmp->key = alloc_name(key, &len);
  if (! len) {
    ERROR("out of memory for key");
    cleanup_entity_edit_sets(tmp);
    return 4;
  }
  TRACE("tmp->key: %s\n", tmp->key);

  /* value */
  if (value) {
    if (options & EDIT_EDIT || options & EDIT_ADD) {
      if (*value) {
        tmp->value = alloc_name(value, &len);
        if (! len) {
          ERROR("out of memory for value");
          cleanup_entity_edit_sets(tmp);
          return 5;
        }
        TRACE("tmp->value: %s\n", tmp->value);
      }
      else if (options & EDIT_DELETE) {
        TRACE("tmp->value: \"\" - delete this key\n");
      }
    }
  }
  else if (! (options & EDIT_DELETE)) {
    ERROR("blank value name\n");
    cleanup_entity_edit_sets(tmp);
    return 6;
  }

  /* options */
  tmp->options = options;
  TRACE("tmp->options: %d\n", tmp->options);

  last = (entity_edit_set_t *) gll_last((gll_t *) *sets);
  TRACE("last: %p\n", last);
  /* this is the first one */
  if (last) last->next = tmp;
  else *sets = tmp;
  /*gll_debug((gll_t *) *sets);*/

  return 0;
}

int add_entities(scn_t *scn) {
  entity_add_set_t *set;
  int ret;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  SAY("adding entities\n");
  GLL_LOOP(set, scn->option->entity_adds) {
    ret = add_entity(scn, set);
    if (ret) {
      ERROR("add_entity() returned %d\n", ret);
      return 2;
    }
  }

  return 0;
}

int add_entity(scn_t *scn, entity_add_set_t *set) {
  int i;
  unsigned short entity_size;
  unsigned short len;
  pair_t *pair;
  thing_t *entities;
  entity_t *entity, **new_things;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }
  if (! set) {
    ERROR("no set!\n");
    return 2;
  }

  entities = &scn->entities;

  /* create a new entity with ClassName and TargetName attributes */
  entity = alloc_entity();
  entity->num_fields = 2;
  entity->fields = alloc_pairs(entity->num_fields);
  if (! entity->fields) {
    ERROR("out of memory for new entity data");
    return 3;
  }

  entity->id = entities->num;
  entity->unknown = 0xffffffff;
  entity_size = sizeof(entity->num_fields) + sizeof(entity->unknown);

  /* ClassName */
  pair = entity->fields[0];
  pair->key = alloc_name(CLASSNAME, &len);
  if (! len) {
    ERROR("out of memory for new entity %s attribute name", CLASSNAME);
    cleanup_entity(entity);
    return 4;
  }
  pair->key_length = len + 1; /* include \0 */

  pair->value = alloc_name(set->classname, &len);
  if (! len) {
    ERROR("out of memory for new entity %s attribute value", CLASSNAME);
    cleanup_entity(entity);
    return 5;
  }
  pair->value_length = len + 1; /* include \0 */

  entity_size += pair->key_length + pair->value_length;
  entity_size += sizeof(pair->key_length) + sizeof(pair->value_length);

  /* TargetName */
  pair = entity->fields[1];
  pair->key = alloc_name(TARGETNAME, &len);
  if (! len) {
    ERROR("out of memory for new entity %s attribute key", TARGETNAME);
    cleanup_entity(entity);
    return 6;
  }
  pair->key_length = len + 1; /* include \0 */

  pair->value = alloc_name(set->name, &len);
  if (! len) {
    ERROR("out of memory for new entity %s attribute value", TARGETNAME);
    cleanup_entity(entity);
    return 7;
  }
  pair->value_length = len + 1; /* include \0 */

  entity_size += pair->key_length + pair->value_length;
  entity_size += sizeof(pair->key_length) + sizeof(pair->value_length);

  /* add entity to list */
  new_things = alloc_entities(entities->num + 1);
  if (! new_things) {
    ERROR("out of memory for new entities");
    return 8;
  }

  /* add existing */
  for (i = 0; i < entities->num; i++) {
    new_things[i] = entities->things[i];
  }

  /* add new */
  new_things[entities->num] = entity;
  free(entities->things);
  entities->num++;
  entities->things = (void **) new_things;

  /* keep track of size */
  entities->length += entity_size;

  SAY("added entity %u (%s=\"%s\"; %s=\"%s\")\n", entity->id, CLASSNAME, set->classname, TARGETNAME, set->name);

  return 0;
}

int delete_entities(scn_t *scn) {
  entity_delete_set_t *set;
  int ret;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  SAY("deleting entities\n");
  GLL_LOOP(set, scn->option->entity_deletes) {
    ret = delete_entity(scn, set);
    if (ret) {
      ERROR("delete_entity() returned %d\n", ret);
      return 2;
    }
  }

  return 0;
}

int delete_entity(scn_t *scn, entity_delete_set_t *set) {
  int i;
  unsigned int j;
  unsigned int entity_size;
  thing_t *entities;
  entity_t *entity, **new_things;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }
  if (! set) {
    ERROR("no set!\n");
    return 2;
  }

  entities = &scn->entities;

  /* sanity */
  if (set->entity_id >= entities->num) {
    ERROR("entity ID %u out of range (%u entit%s)\n", set->entity_id, entities->num, (entities->num == 1) ? "y" : "ies");
    return 3;
  }

  /* find the size of this entity to subtract from entities->length */
  entity_size = 0;
  entity = entities->things[set->entity_id];

  entity_size += sizeof(entity->num_fields);
  entity_size += sizeof(entity->unknown);
  for (j = 0; j < entity->num_fields; j++) {
    pair_t *pair = entity->fields[j];
    entity_size += pair->key_length + pair->value_length;
    entity_size += sizeof(pair->key_length) + sizeof(pair->value_length);
  }

  SAY("deleting entity %u with size %u\n", set->entity_id, entity_size);

  new_things = alloc_entities(entities->num - 1);
  if (! new_things) {
    ERROR("out of memory for new entities");
    return 4;
  }

  /* add entities before the victim */
  for (i = 0; i < set->entity_id; i++) {
    new_things[i] = entities->things[i];
  }

  /* add entities after the victim */
  for (i = set->entity_id + 1; i < entities->num; i++) {
    new_things[i - 1] = entities->things[i];
  }

  /* replace */
  free(entities->things);
  entities->num--;
  entities->length -= entity_size;
  entities->things = (void **) new_things;

  return 0;
}

int edit_entities(scn_t *scn) {
  entity_edit_set_t *set;
  int ret;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  SAY("editing entities\n");
  GLL_LOOP(set, scn->option->entity_edits) {
    ret = edit_entity(scn, set);
    if (ret) {
      ERROR("edit_entity() returned %d\n", ret);
      return 2;
    }
  }

  return 0;
}

int edit_entity(scn_t *scn, entity_edit_set_t *set) {
  unsigned int i, j;
  unsigned short len;
  pair_t *pair, *new_pair, **new_pairs;
  thing_t *entities;
  entity_t *entity;
  char *string;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }
  if (! set) {
    ERROR("no set!\n");
    return 2;
  }

  entities = &scn->entities;

  /* sanity */
  if (set->entity_id >= entities->num) {
    ERROR("entity ID %u out of range (%u entit%s)\n", set->entity_id, entities->num, (entities->num == 1) ? "y" : "ies");
    return 3;
  }

  entity = entities->things[set->entity_id];

  /* find the pair we're interested in */
  pair = 0;
  for (i = 0; i < entity->num_fields; i++) {
    new_pair = entity->fields[i];
    if (str_equiv(new_pair->key, set->key)) {
      pair = new_pair;
      break;
    }
    /* don't reuse i; we remember this pair ID when deleting */
  }

  /* create a new pair */
  if (! pair) {
    if (set->options & EDIT_DELETE) {
      error_dbg("entity %u doesn't have an attribute named \"%s\"\n", set->entity_id, set->key);
      return 0;
    }

    if (! set->value) {
      ERROR("blank value for new key \"%s\"\n", set->key);
      return 4;
    }
    if (! (set->options & EDIT_ADD)) {
      ERROR("attribute \"%s\" doesn't exist in entity %u\n", set->key, set->entity_id);
      return 5;
    }

    SAY("creating new attribute \"%s\" with value \"%s\" for entity %u\n", set->key, set->value, set->entity_id);
    /* XXX: move me into a function? */
    new_pairs = alloc_more_pairs(1, entity->num_fields, entity->fields);
    if (! new_pairs) {
      ERROR("out of memory for new field");
      return 6;
    }
    entity->fields = new_pairs;
    pair = entity->fields[entity->num_fields++];

    /* set key for new pair */
    pair->key = alloc_name(set->key, &len);
    if (! len) {
      ERROR("out of memory for new key \"%s\"");
      return 7;
    }
    pair->key_length = len + 1; /* include \0 */

    /* set value for new pair */
    pair->value = alloc_name(set->value, &len);
    if (! len) {
      ERROR("out of memory for new value \"%s\"");
      return 7;
    }
    pair->value_length = len + 1; /* include \0 */

    /* adjust data stream length */
    entities->length += pair->key_length + pair->value_length;
    entities->length += sizeof(pair->key_length) + sizeof(pair->value_length);

    return 0;
  }

  if (! set->value) {
    /* delete pair */
    SAY("deleting entity %u attribute \"%s\" (was \"%s\")\n", set->entity_id, pair->key, pair->value);

    /* adjust data stream length */
    entities->length -= (pair->key_length + pair->value_length + 4);

    cleanup_pair(pair);
    for (j = i; j < entity->num_fields - 1; j++) {
      entity->fields[j] = entity->fields[j + 1];
    }
    entity->num_fields--;

    if (str_equiv(CLASSNAME, set->key)) {
      WARN("deleting entity %u's %s attribute was pretty stupid\n", entity->id, CLASSNAME);
    }

    return 0;
  }

  /* edit an existing pair */
  if (! (set->options & EDIT_EDIT)) {
    ERROR("attribute \"%s\" already exists in entity %u\n", set->key, set->entity_id);
    return 8;
  }

  /* no-op */
  if (str_equals(pair->value, set->value)) {
    error_dbg("entity %u attribute \"%s\" is already set to \"%s\"\n", set->entity_id, set->key, set->value);
    return 0;
  }

  SAY("setting value of entity %u attribute \"%s\" to \"%s\" (was \"%s\")\n", set->entity_id, set->key, set->value, pair->value);
  string = alloc_name(set->value, &len);
  if (! len) {
    ERROR("out of memory for value \"%s\"", set->value);
    return 9;
  }
  free(set->value);

  /* adjust data stream length */
  entities->length += len + 1 - pair->value_length;

  pair->value = string;
  pair->value_length = len + 1; /* include \0 */

  return 0;
}

int summarise_entities(scn_t *scn) {
  thing_t *entities;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  entities = &scn->entities;

  SAY("Map entities: %u\n", entities->num);

  return 0;
}

int describe_entities(scn_t *scn) {
  int i;
  thing_t *entities;
  entity_t *entity;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  entities = &scn->entities;

  SAY("Map entities: %u at 0x%08x, length 0x%04x\n", entities->num, entities->offset, entities->length);
  for (i = 0; i < entities->num; i++) {
    entity = (entity_t *) entities->things[i];
    if (describe_entity(entity, 2)) return 2;
  }

  return 0;
}

int describe_entity(entity_t *entity, int indent) {
  unsigned int i;
  int j;
  pair_t *pair;

  if (! entity) {
    ERROR("no entity!\n");
    return 1;
  }

  SAY("Entity %u at 0x%08x, fields: %u\n", entity->id, entity->offset, entity->num_fields);
  for (i = 0; i < entity->num_fields; i++) {
    pair = entity->fields[i];
    for (j = 0; j < indent; j++) SAY(" ");
    SAY("0x%08x: %s -> \"%s\"\n", pair->offset, pair->key, pair->value);
  }

  return 0;
}
