#include "scnhacking.h"

/* open the file for read-only access */
FILE *open_read(char *filename) {
  FILE *fd = fopen(filename, "rb");
  TRACE("fopen({filename=\"%s\"}, \"rb\"): {fd=%p}\n", filename, fd);
  if (! fd) ERROR("error opening file for reading");
  return fd;
}

/* open the file for write-only access */
FILE *open_write(char *filename) {
  FILE *fd = fopen(filename, "wb");
  TRACE("open({filename=\"%s\"}, \"wb\"): {fd=%p}\n", filename, fd);
  if (! fd) ERROR("error opening file for writing");
  return fd;
}

/* find our position in the file by seeking 0 octets */
off_t get_file_position(FILE *fd) {
  off_t ret = ftell(fd);
  if (ret == -1) ERROR("can't find current file position");
  return ret;
}

/* skip unknown stuff */
int skip(FILE *fd, off_t length) {
  int ret = fseek(fd, length, SEEK_CUR);
  TRACE1("skip({fd=%p}, %d): 0x%04x\n", fd, length, ret);
  if (ret) {
    ERROR("can't seek past unknown information");
    return -1;
  }
  return ret;
}

/* try a few times to read stuff into our buffer */
size_t try_read(FILE *fd, void *dest, size_t count) {
  void *buf;
  size_t in;
  int tries;

  if (! fd) {
    ERROR("null file pointer\n");
    return -1;
  }

  buf = dest; tries = 0;
  while (tries < 5) {
    in = fread(buf, 1, count, fd);
    TRACE1("fread({buf=%p}, 1, {count=%u}, {fd=%p}): %u (expected %u)\n", buf, count, fd, in, count);
    if (in == count) return in;
    else if (ferror(fd)) {
      clearerr(fd);
      tries++;
      DEBUG("failed read on fd %p (try %d)\n", fd, tries);
      count -= in;
#ifdef _MSC_VER
      (char *) buf += in;
#else
      buf += in;
#endif
    }
    else {
      TRACE("nothing read, zeroing buf\n");
      memset(dest, 0, count);
      return 0;
    }
  }

  ERROR("multiple failed reads on fd %p, buffer %p is dirty\n", fd, dest);
  return -1;
}

/* read in a char */
int read_char(FILE *fd, char *buf) {
  switch (try_read(fd, buf, sizeof(char))) {
     case 0: return 1;
     case sizeof(char): return 0;
  }
  return -1;
}

/* read in an unsigned char */
int read_unsigned_char(FILE *fd, unsigned char *buf) {
  switch (try_read(fd, buf, sizeof(char))) {
     case 0: return 1;
     case sizeof(char): return 0;
  }
  return -1;
}

/* read in a short */
int read_uint16(FILE *fd, unsigned short *buf) {
  switch (try_read(fd, buf, sizeof(short))) {
     case 0: return 1;
     case sizeof(short): return 0;
  }
  return -1;
}

/* read in an int */
int read_uint32(FILE *fd, unsigned int *buf) {
  switch (try_read(fd, buf, sizeof(int))) {
     case 0: return 1;
     case sizeof(int): return 0;
  }
  return -1;
}

/* read in a float */
int read_float(FILE *fd, float *buf) {
  switch (try_read(fd, buf, sizeof(float))) {
     case 0: return 1;
     case sizeof(float): return 0;
  }
  return -1;
}

/* read an arbitrary-length buffer */
int read_arbitrary(FILE *fd, void *buf, unsigned int len) {
  size_t in = try_read(fd, buf, len);
  if (in == len) return 0;
  else if (in) return -1;
  return 1;
}

/* XXX: unreliable stdio */
/* XXX: Unused in scnhacking so ignore signed/unsigned warnings. */
int read_line(FILE *fd, char *buffer, size_t len) {
  if (fgets(buffer, len, fd)) return 0;
  if (feof(fd)) {
    DEBUG("unexpected end of file on fd %p\n", fd);
    return 1;
  }
  ERROR("read failed on fd %p\n", fd);
  return -1;
}

/* try a few times to write stuff from our buffer */
size_t try_write(FILE *fd, void *src, size_t count) {
  void *buf;
  size_t out;
  int tries;

  if (! fd) {
    ERROR("null file pointer\n");
    return -1;
  }

  buf = src; tries = 0;
  while (tries < 5) {
    out = fwrite(buf, 1, count, fd);
    TRACE1("write({buf=%p}, 1, {count=%u}, {fd=%p}): %u (expected %u)\n", buf, count, fd, out, count);
    if (out == count) return out;
    else if (ferror(fd)) {
      tries++;
      DEBUG("failed write on fd %p (try %d)\n", fd, tries);
      count -= out;
#ifdef _MSC_VER
      (char *) buf += out;
#else
      buf += out;
#endif
    }
  }

  ERROR("multiple failed writes on fd %p\n", fd);
  return -1;
}

/* write out a char */
int write_char(FILE *fd, char *buf) {
  switch (try_write(fd, buf, sizeof(char))) {
     case 0: return 1;
     case sizeof(char): return 0;
  }
  return -1;
}

/* write out an unsigned char */
int write_unsigned_char(FILE *fd, unsigned char *buf) {
  switch (try_write(fd, buf, sizeof(char))) {
     case 0: return 1;
     case sizeof(char): return 0;
  }
  return -1;
}

/* write out a short */
int write_uint16(FILE *fd, unsigned short *buf) {
  switch (try_write(fd, buf, sizeof(short))) {
     case 0: return 1;
     case sizeof(short): return 0;
  }
  return -1;
}

/* write out an int */
int write_uint32(FILE *fd, unsigned int *buf) {
  switch (try_write(fd, buf, sizeof(int))) {
     case 0: return 1;
     case sizeof(int): return 0;
  }
  return -1;
}

/* right out a float */
int write_float(FILE *fd, float *buf) {
  switch (try_write(fd, buf, sizeof(float))) {
     case 0: return 1;
     case sizeof(float): return 0;
  }
  return -1;
}

/* write an arbitrary-length buffer */
int write_arbitrary(FILE *fd, void *buf, unsigned int len) {
  size_t in = try_write(fd, buf, len);
  if (in == len) return 0;
  else if (in) return -1;
  return 1;
}

/* XXX: unreliable stdio */
int write_string(FILE *fd, char *format, ...) {
  va_list args;

  va_start(args, format);
  vfprintf(fd, format, args);
  va_end(args);

  return 0;
}

/* check that a (texture) filename is valid on Windows */
int invalid_windows_filename(unsigned char *buf, unsigned short len) {
  unsigned short i;
  int seen_colon, seen_backslash;
  
  seen_colon = seen_backslash = 0;

  for (i = 0; i < len; i++) {
    if (! isprint(buf[i])) return 1;
    /* allowed in drive assignment */
    if (buf[i] == ':') if (seen_colon++) return 1;
    if (buf[i] == '/') return 1;
    if (buf[i] == '*') return 1;
    if (buf[i] == '?') return 1;
    if (buf[i] == '"') return 1;
    if (buf[i] == '<') return 1;
    if (buf[i] == '>') return 1;
    if (buf[i] == '|') return 1;
  }

  return 0;
}

/* check that a texture file was found */
int missing_file(char *filename, unsigned short len) {
#ifdef _MSC_VER
  return 0;
#else
  struct dirent *dir;
  struct stat st;
  char buf[65536], *cwd;
  int ret;
  DIR *directory = opendir(".");

  TRACE("opendir(\".\"): {directory=%p}\n", directory);
  cwd = getcwd(buf, sizeof(buf));
  if (! cwd) {
    ERROR("can't get current directory");
    closedir(directory);
    return 1;
  }
  DEBUG("looking for \"%s\" in %s\n", filename, cwd);

  if (! directory) {
    ERROR("can't check for texture in current directory %s", cwd);
    closedir(directory);
    return 2;
  }

  while (dir = readdir(directory)) {
    TRACE("readdir({directory=%p}): {dir=%p}\n", directory, dir->d_name);
    if (str_diff(dir->d_name, ".") && str_diff(dir->d_name, "..")) {
      if (stat(dir->d_name, &st)) {
        ERROR("can't stat directory %s", dir->d_name);
        TRACE("closedir({directory=%p})\n", directory);
        closedir(directory);
        return 3;
      }
      if (S_ISDIR(st.st_mode)) {
        TRACE("\"%s\" is a directory\n", dir->d_name);
        ret = chdir(dir->d_name);
        TRACE("chdir(\"%s\"): %d\n", dir->d_name, ret);
        if (ret) {
          ERROR("unable to chdir(\"%s\")", dir->d_name);
          continue;
        }
        if (! missing_file(filename, len)) {
          TRACE("closedir({directory=%p})\n", directory);
          closedir(directory);
          return 0;
        }
        ret = chdir(cwd);
        TRACE("chdir(\"%s\"): %d\n", cwd, ret);
        if (ret) {
          ERROR("unable to chdir(\"%s\") - aborting\n", cwd);
          closedir(directory);
          return 0;
        }
      }
      /* let's assume it's a file */
      else {
        TRACE("checking filename %s\n", dir->d_name);
        if (str_equiv(dir->d_name, filename)) {
          { DEBUG("found %s in %s\n", filename, cwd); }
          TRACE("closedir({directory=%p})\n", directory);
          closedir(directory);
          return 0;
        }
      }
    }
    else TRACE("skipping \"%s\"\n", dir->d_name);
  }
  if (! dir) ERROR("can't read current directory");

  TRACE("closedir({directory=%p})\n", directory);
  closedir(directory);
  return 4;
#endif
}

/* find basename and put it in dest pointer*/
void set_basename(char *filename, unsigned short filename_length, char **basename, unsigned short *basename_length) {
  size_t i;

  /* hmm, do I look like a Sierra employee? */
  if (! basename) {
    ERROR("called with null basename pointer\n");
    return;
  }

  TRACE1("basename: %p, basename_length: %p\n", basename, basename_length);
  if (! filename_length) filename_length = (unsigned short) strlen(filename);
  for (i = 0; i < filename_length; i++) {
    if (filename[filename_length - i] == '\\' || filename[filename_length - i] == '/') {
      i--;
      break;
    }
  }
  if (basename_length) *basename_length = (unsigned short) i;
  *basename = filename + filename_length - i;
}

/* ignore milkshape style // comments */
void ignore_comments(char *line) {
  char *s;

  for (s = line; *s; s++) {
    if (*s == '/' && *(s + 1) == '/') {
      if (s == line) continue;
      if (*(s - 1) == '\\') continue;
      *s = '\0';
      return;
    }
    if (*s == '\r' || *s == '\n') {
      *s = '\0';
      return;
    }
  }
}

/* $line =~ /^\s*$/ */
int null_line(char *line) {
  char *s;

  for (s = line; *s; s++) if (! isspace(*s)) return 0;
  return 1;
}

#ifndef _MSC_VER
/* XXX: unreliable stdio */
int read_string(char *buf, char *format, ...) {
  va_list args;
  int in = 0;

  va_start(args, format);
  in += vsscanf(buf, format, args);
  va_end(args);

  return in;
}
#endif


