#include "scnhacking.h"

int main(int, char **);

int quiet = 0;
int debug = 0;
int verbose = 0;
int num_files;
FILE *error;

int main(int argc, char **argv) {
  int i;
  scn_t **scns, *scn;
  struct stat st;
  int errors, need_write, need_light_recount, ret;

#ifdef _MSC_VER
  error = stdout;
#else
  error = stderr;
#endif

  scns = process_options(&argc, argv, &num_files);
  if (! scns) exit(1);
  if (num_files == -1) exit(2);
  DEBUG("num_files: %d\n", num_files);

  /* first check we can open all files (not => args error) */
  for (i = 0; i < num_files; i++) {
    if (stat(scns[i]->filename, &st)) {
      ERROR("can't stat file %s", scns[i]->filename);
      exit(2);
    }
#ifndef _MSC_VER
    if (S_ISDIR(st.st_mode)) {
      ERROR("%s is a directory\n", scns[i]->filename);
      exit(2);
    }
#endif
  }

  errors = 0;
  for (i = 0; i < num_files; i++) {
    scn = scns[i];

    if (scn->option->fd == stdout && (debug || verbose)) {
      ERROR("can't write to stdout in debug or verbose mode\n");
      errors++;
      continue;
    }

    need_write = 0;
    need_light_recount = 0;

    /* open file */
    if (read_scn_file(scn)) {
      errors++;
      goto cleanup_no_close;
    }

    /* parse file */
    if (parse_scn(scn)) {
      errors++;
      goto cleanup;
    }

    /* skip back to start */
    if (fseek(scn->fd, 0, SEEK_SET)) {
      ERROR("can't seek back to start of %s\n", scn->filename);
      errors++;
      goto cleanup;
    }

    if (! scn->option->number) {
      scn->option->describe = DESCRIBE_SUMMARY;
    }

    /* remove lightmaps from end of file - use "swat.exe -blmaps" to generate */
    if (scn->option->strip_lightmaps) {
      ret = strip_lightmaps(scn);
      if (ret) {
        ERROR("lightmap stripping function returned %d\n", ret);
        errors++;
        goto cleanup;
      }
      else need_write = 1;
    }

    /* replace colours in the file */
    if (scn->option->colour_replaces) {
      ret = replace_colours(scn);
      if (ret) {
        ERROR("colour replacement function returned %d\n", ret);
        errors++;
      }
      else need_write = 1;
    }

    /* add entities */
    if (scn->option->entity_adds) {
      ret = add_entities(scn);
      if (ret) {
        ERROR("entity add function returned %d\n", ret);
        errors++;
      }
      else need_write = need_light_recount = 1;
    }

    /* delete entities */
    if (scn->option->entity_deletes) {
      ret = delete_entities(scn);
      if (ret) {
        ERROR("entity delete function returned %d\n", ret);
        errors++;
      }
      else need_write = need_light_recount = 1;
    }

    /* edit entities */
    if (scn->option->entity_edits) {
      ret = edit_entities(scn);
      if (ret) {
        ERROR("entity edit function returned %d\n", ret);
        errors++;
      }
      else need_write = need_light_recount = 1;
    }

    if (need_light_recount) {
      if (count_lights(scn)) {
        need_write = 0;
        errors++;
      }
    }

    /* list entities */
    if (scn->option->entity_searches) {
      if (search_entities(scn)) errors++;
    }

    /* check out SCN file information */
    if (scn->option->describe) {
      if (scn->option->describe & DESCRIBE_SUMMARY) {
        if (summarise_scn(scn)) {
          errors++;
          goto cleanup;
        }
      }
      if (scn->option->describe & DESCRIBE_STANDARD) {
        if (describe_scn(scn)) {
          errors++;
          goto cleanup;
        }
      }
    }

    if (need_write) {
      errors++;
      if (write_scn_file(scn)) goto cleanup;
      fclose(scn->fd);
      if (link_scn_file(scn)) goto cleanup;
      errors--;
      goto cleanup_no_close;
    }

cleanup:
    fclose(scn->fd);
cleanup_no_close:
    cleanup_scn(scn);
  }
  free(scns);

  exit(errors);
}
