#include "scnhacking.h"

scn_t **alloc_scns(unsigned int n) {
  return (struct scn **) alloc_lots(n, "scn", sizeof(struct scn *), (void *(*)()) alloc_scn);
}

scn_t *alloc_scn() {
  return (struct scn *) alloc_one("scn", sizeof(struct scn));
}

void cleanup_scn(scn_t *scn) {
  if (! scn) {
    ERROR("no SCN!\n");
    return;
  }
  if (scn->filename) free(scn->filename);
  if (scn->outfile) free(scn->outfile);
  if (scn->tmpfile) free(scn->tmpfile);
  cleanup_entities(&scn->entities);
  if (scn->lightmaps.num_lights) free(scn->lightmaps.addresses);
  cleanup_entclasses(scn->num_entclasses, scn->entclasses);
  cleanup_option(scn->option);
  TRACE2("freeing SCN at %p\n", scn);
  free(scn);
}

int read_scn_file(scn_t *scn) {
  if (! scn || ! scn->filename) {
    ERROR("no SCN!\n");
    return 1;
  }

  scn->fd = open_read(scn->filename);
  if (! scn->fd) {
    ERROR("can't open %s\n", scn->filename);
    return 2;
  }
  DEBUG("opened %s on descriptor %p\n", scn->filename, scn->fd);

  if (check_scn_magic(scn)) {
    fclose(scn->fd);
    return 3;
  }
  set_basename(scn->filename, 0, &scn->basename, &scn->basename_length);

  return 0;
}

int check_scn_magic(scn_t *scn) {
  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  /* read magic */
  if (read_arbitrary(scn->fd, &scn->magic, sizeof(scn->magic))) {
    ERROR("can't read %s magic\n", scn->filename);
    return 2;
  }

  /* check magic */
  if (memcmp(&scn->magic, SCN_MAGIC, sizeof(SCN_MAGIC_LENGTH))) {
    ERROR("%s isn't an SCN file\n", scn->filename);
    return 3;
  }

  return 0;
}

int parse_scn(scn_t *scn) {
  int buf;
  off_t pos;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  /* DEBUGGING: move me to a function or delete me */
  do {
    pos = get_file_position(scn->fd);
    if (read_uint32(scn->fd, &buf)) return -1;
    if (buf) DEBUG("0x%08x: 0x%08x (%u)\n", pos, buf, buf);
  } while (pos < 0xc0);

  /* move this into one function once we know what all the values mean */
  if (find_num_entities(scn)) return 2;
  if (find_lightmaps(scn)) return 3;
  if (find_entities(scn)) return 4;
  if (parse_lightmaps(scn)) return 5;

  return 0;
}

int write_scn_file(scn_t *scn) {
  if (! scn || (! scn->filename && ! scn->option->fd)) {
    ERROR("no SCN to write!\n");
    return 1;
  }

  if (open_scn_file(scn)) return 2;
  if (write_scn_data(scn)) return 3;

  return 0;
}

int open_scn_file(struct scn *scn) {
  char *dirname;
  unsigned short dirname_length;
  int need_free;

  if (! scn || (! scn->filename && ! scn->option->fd)) {
    ERROR("no SCN to write!\n");
    return 1;
  }

  /* write to stdout? */
  if (scn->option->fd) scn->out_fd = scn->option->fd;
  else {
    /* use existing filename if no output filename was given */
    scn->outfile = scn->option->outfile;
    if (! scn->outfile) {
      scn->outfile = scn->filename;
      /* was there a file prefix? */
      if (scn->option->prefix) scn->outfile = add_prefix(scn->outfile, scn->option->prefix);
      if (scn->option->suffix) scn->outfile = add_suffix(scn->outfile, scn->option->suffix);
      if (! scn->outfile) {
        ERROR("scn->outfile got mashed\n");
        return 2;
      }
    }
    set_basename(scn->outfile, 0, &scn->out_basename, &scn->out_basename_length);
    /* create tmp file */
    scn->tmpfile = (char *) malloc(sizeof(char) * 24 + strlen(scn->outfile));
    if (! scn->tmpfile) {
      ERROR("out of memory for scn temporary output file");
      return 3;
    }

    dirname = get_dirname(scn->outfile, &dirname_length, &need_free);
    if (! dirname) {
      ERROR("can't get outfile dirname\n");
      return 4;
    }

    DEBUG("dirname_length: %u\n", dirname_length);
    if (dirname_length) {
      *(dirname + dirname_length - 1) = PATHSEP;
      *(dirname + dirname_length) = '\0';
    }
    else {
      /*free(dirname);*/ /* not needed? */
      need_free = 0;
      dirname = "";
    }
    if (snprintf(scn->tmpfile, 24 + strlen(scn->outfile), "%s%u.%u.%s", dirname, (unsigned int) time(0), getpid(), scn->out_basename) == -1) {
      ERROR("snprintf() failed writing scratch file\n");
      return 5;
    }
    if (need_free) free(dirname);

    scn->out_fd = open_write(scn->tmpfile);
    if (! scn->out_fd) {
      ERROR("can't open tmpfile %s!\n", scn->tmpfile);
      return 6;
    }
    DEBUG("opened tmpfile %s on descriptor %p\n", scn->tmpfile, scn->out_fd);
  }

  return 0;
}

int link_scn_file(struct scn *scn) {
  if (! scn || ! scn->out_fd) {
    ERROR("SCN has no open file descriptor\n");
    return 1;
  }

  if (scn->option->fd);
  else {
    DEBUG("closing descriptor %p for %s\n", scn->out_fd, scn->tmpfile);
    if (fclose(scn->out_fd)) {
      ERROR("failed to close file descriptor %p\n", scn->out_fd);
      return 2;
    }
    scn->out_fd = 0;
  
    DEBUG("linking real output filename %s\n", scn->outfile);
#ifdef _MSC_VER
    if (str_equals(scn->filename, scn->outfile)) {
      if (unlink(scn->outfile)) {
        ERROR("failed to unlink original file %s prior to replacement\n", scn->outfile);
      }
    }
#endif
    if (rename(scn->tmpfile, scn->outfile)) {
      ERROR("failed to create output file %s\n", scn->outfile);
      ERROR("SCN data can be found in temp file %s\n", scn->tmpfile);
      return 3;
    }
  
    INFO("%s written successfully\n", scn->outfile);
  }
  if (scn->outfile == scn->filename) scn->outfile = 0;
  return 0;
}

int close_scn_file(struct scn *scn) {
  int ret;

  if (! scn || ! scn->fd) {
    ERROR("SCN has no open file descriptor\n");
    return -1;
  }
  DEBUG("closing descriptor %p for %s\n", scn->fd, scn->filename);
  ret = fclose(scn->fd);
  scn->fd = 0;

  return ret;
}

/* XXX: signed/unsigned problems with off_t, size_t and int here */
int write_scn_data(scn_t *scn) {
  char buf[4096];
  off_t pos, endpos;
  int count, size, finished;
  thing_t *entities;

  if (! scn || ! scn->fd || ! scn->out_fd) {
    ERROR("no SCN!\n");
    return 1;
  }

  entities = &scn->entities;

  /* read and write up until entities */
  count = finished = 0;
  while (! finished ){
    finished = 1;

    size = scn->entities.offset - count;
    if (size > sizeof(buf)) size = sizeof(buf);
    if (read_arbitrary(scn->fd, &buf, size)) {
      ERROR("can't slurp %s\n", scn->filename);
      return 2;
    }
    if (feof(scn->fd)) {
      ERROR("unexpected end of file\n");
      return 3;
    }
    count += size;
    if (write_arbitrary(scn->out_fd, &buf, size)) {
      ERROR("can't write %s\n", scn->outfile);
      return 4;
    }

    if (count < scn->entities.offset) finished = 0;
  }

  pos = get_file_position(scn->out_fd);
  TRACE("wrote entities: now at 0x%08x (expected 0x%08x)\n", pos, scn->entities.offset);
  if (write_scn_header(scn)) return 5;
  if (fseek(scn->out_fd, pos, SEEK_SET)) {
    ERROR("can't seek to write entities");
    return 6;
  }
  if (write_entities(scn)) return 7;

  /* write lightmaps offset */
  pos = get_file_position(scn->out_fd);
  DEBUG("lightmaps offset was 0x%08x and now is 0x%08x\n", scn->lightmaps.offset, pos);
  scn->lightmaps.offset = pos;
  if (fseek(scn->out_fd, LIGHTMAPS_OFFSET, SEEK_SET)) {
    ERROR("can't seek to write lightmaps offset");
    return 7;
  }
  if (write_uint32(scn->out_fd, (unsigned int *) &scn->lightmaps.offset)) {
    ERROR("can't write lightmaps offset\n");
    return 8;
  }
  /* write number of lights */
  if (write_uint32(scn->out_fd, &scn->lightmaps.num_lights)) {
    ERROR("can't write number of lights\n");
    return 9;
  }

  /* write lightmaps */
  if (scn->option->strip_lightmaps) return 0;

  /* skip past entities */
  count += entities->orig_length;

  /* seek to end of file */
  if (fseek(scn->fd, 0, SEEK_END)) {
    ERROR("can't find end of file");
    return 10;
  }
  /* remember end of file position */
  endpos = get_file_position(scn->fd);
  /* go back to old positions */
  if (fseek(scn->fd, count, SEEK_SET)) {
    ERROR("can't seek back on input descriptor");
    return 11;
  }
  if (fseek(scn->out_fd, scn->lightmaps.offset, SEEK_SET)) {
    ERROR("can't seek back on output descriptor");
    return 11;
  }

  finished = 0;
  while (! finished) {
    finished = 1;

    size = endpos - count;
    if (size > sizeof(buf)) size = sizeof(buf);
    if (read_arbitrary(scn->fd, &buf, size)) {
      ERROR("can't read %s\n", scn->filename);
      return 12;
    }
    if (write_arbitrary(scn->out_fd, &buf, size)) {
      ERROR("can't write %s\n", scn->outfile);
      return 13;
    }
    if (feof(scn->fd)) continue;
    count += size;

    if (count < endpos) finished = 0;
  }

  return 0;
}

/* XXX: big hack function */
int write_scn_header(scn_t *scn) {
  thing_t *entities;

  if (! scn || ! scn->out_fd) {
    ERROR("no SCN!\n");
    return 1;
  }

  entities = &scn->entities;

  /* assuming first batch of stuff has been written, overwrite changed data */
  if (fseek(scn->out_fd, ENTITIES_NUM_OFFSET, SEEK_SET)) {
    ERROR("can't seek to number of entities offset");
    return 2;
  }
  if (write_uint32(scn->out_fd, &entities->num)) {
    ERROR("can't write number of entities\n");
    return 3;
  }

  /* length */
  if (fseek(scn->out_fd, ENTITIES_START_OFFSET, SEEK_SET)) {
    ERROR("can't seek to entities start offset");
    return 4;
  }
  if (write_uint32(scn->out_fd, (unsigned int *) &entities->offset)) {
    ERROR("can't write entities offset\n");
    return 5;
  }
  if (write_uint32(scn->out_fd, (unsigned int *) &entities->length)) {
    ERROR("can't write entities data length\n");
    return 6;
  }

  /* write offset of lights? */
  if (fseek(scn->out_fd, LIGHTMAPS_OFFSET, SEEK_SET)) {
    ERROR("can't seek to lightmaps position offset");
    return 5;
  }

  entities->offset += entities->length - entities->orig_length;
  DEBUG("writing offset 0x%08x from length %u\n", entities->offset, entities->length);
  if (write_uint32(scn->out_fd, (unsigned int *) &entities->offset)) {
    ERROR("can't write lightmaps offset\n");
    return 6;
  }

  return 0;
}

int summarise_scn(scn_t *scn) {
  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  SAY("Summary of %s\n", scn->basename);
  if (summarise_entities(scn)) return 2;
  if (summarise_entclasses(scn)) return 3;
  if (summarise_lightmaps(scn)) return 4;
  if (summarise_lights(scn)) return 5;

  return 0;
}

int describe_scn(scn_t *scn) {
  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  SAY("SCN file: %s\n", scn->basename);
  if (scn->option->describe & DESCRIBE_ENTITIES) if (describe_entities(scn)) return 2;
  if (scn->option->describe & DESCRIBE_LIGHTS) if (describe_lightmaps(scn)) return 3;
  if (scn->option->describe & DESCRIBE_LIGHTS) if (describe_lights(scn)) return 4;
  if (scn->option->describe & DESCRIBE_CLASSES) if (describe_entclasses(scn)) return 5;

  return 0;
}
